'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var slate = require('@gitbook/slate');
var immutable = require('immutable');

function getCurrentItem(opts, value, block) {
    const { document } = value;
    if (!block) {
        if (!value.selection.startKey) {
            return null;
        }
        block = value.startBlock;
    }
    const parent = document.getParent(block.key);
    return parent && parent.type === opts.typeItem ? parent : null;
}

function isList(opts, node) {
    return opts.types.includes(node.type);
}

function getListForItem(opts, value, item) {
    const { document } = value;
    const parent = document.getParent(item.key);
    return parent && isList(opts, parent) ? parent : null;
}

function getCurrentList(opts, value, block) {
    const item = getCurrentItem(opts, value, block);
    if (!item) {
        return null;
    }
    return getListForItem(opts, value, item);
}

function getItemDepth(opts, value, block) {
    const { document, startBlock } = value;
    block = block || startBlock;
    const currentItem = getCurrentItem(opts, value, block);
    if (!currentItem) {
        return 0;
    }
    const list = document.getParent(currentItem.key);
    return 1 + getItemDepth(opts, value, list);
}

function getItemsAtRange(opts, value, range) {
    range = range || value.selection;
    if (!range.startKey) {
        return immutable.List();
    }
    const { document } = value;
    const startBlock = document.getClosestBlock(range.startKey);
    const endBlock = document.getClosestBlock(range.endKey);
    if (startBlock === endBlock) {
        const item = getCurrentItem(opts, value, startBlock);
        return item ? immutable.List([item]) : immutable.List();
    }
    const ancestor = document.getCommonAncestor(startBlock.key, endBlock.key);
    if (isList(opts, ancestor)) {
        const startPath = ancestor.getPath(startBlock.key);
        const endPath = ancestor.getPath(endBlock.key);
        return ancestor.nodes.slice(startPath[0], endPath[0] + 1);
    }
    else if (ancestor.type === opts.typeItem) {
        return immutable.List([ancestor]);
    }
    return immutable.List();
}

function getPreviousItem(opts, value, block) {
    const { document, startBlock } = value;
    block = block || startBlock;
    const currentItem = getCurrentItem(opts, value, block);
    if (!currentItem) {
        return null;
    }
    const previousSibling = document.getPreviousSibling(currentItem.key);
    if (!previousSibling) {
        return null;
    }
    else if (previousSibling.type === opts.typeItem) {
        return previousSibling;
    }
    return null;
}

function isSelectionInList(opts, value, type) {
    const items = getItemsAtRange(opts, value);
    return (!items.isEmpty() &&
        (!type ||
            getListForItem(opts, value, items.first()).get('type') === type));
}

function decreaseItemDepth(opts, change) {
    const { value } = change;
    const { document } = value;
    const depth = getItemDepth(opts, value);
    if (depth === 1) {
        return change;
    }
    const currentItem = getCurrentItem(opts, value);
    if (!currentItem) {
        return change;
    }
    const currentList = document.getParent(currentItem.key);
    const parentItem = document.getParent(currentList.key);
    const parentList = document.getParent(parentItem.key);
    const followingItems = currentList.nodes
        .skipUntil(i => i === currentItem)
        .rest();
    const willEmptyCurrentList = currentList.nodes.size === followingItems.size + 1;
    if (!followingItems.isEmpty()) {
        const sublist = slate.Block.create({
            object: 'block',
            type: currentList.type,
            data: currentList.data
        });
        change.insertNodeByKey(currentItem.key, currentItem.nodes.size, sublist, {
            normalize: false
        });
        change.moveNodeByKey(currentItem.key, parentList.key, parentList.getChildIndex(parentItem.key) + 1, { normalize: false });
        followingItems.forEach((item, index) => change.moveNodeByKey(item.key, sublist.key, sublist.nodes.size + index, {
            normalize: false
        }));
    }
    else {
        change.moveNodeByKey(currentItem.key, parentList.key, parentList.getChildIndex(parentItem.key) + 1);
    }
    if (willEmptyCurrentList) {
        change.removeNodeByKey(currentList.key);
    }
    return change;
}

function increaseItemDepth(opts, change) {
    const previousItem = getPreviousItem(opts, change.value);
    const currentItem = getCurrentItem(opts, change.value);
    if (!previousItem) {
        return change;
    }
    if (!currentItem) {
        return change;
    }
    return moveAsSubItem(opts, change, currentItem, previousItem.key);
}
function moveAsSubItem(opts, change, item, destKey) {
    const destination = change.value.document.getDescendant(destKey);
    const lastIndex = destination.nodes.size;
    const lastChild = destination.nodes.last();
    const existingList = isList(opts, lastChild) ? lastChild : null;
    if (existingList) {
        return change.moveNodeByKey(item.key, existingList.key, existingList.nodes.size);
    }
    const currentList = getListForItem(opts, change.value, destination);
    if (!currentList) {
        throw new Error('Destination is not in a list');
    }
    const newSublist = slate.Block.create({
        object: 'block',
        type: currentList.type,
        data: currentList.data
    });
    change.insertNodeByKey(destKey, lastIndex, newSublist, {
        normalize: false
    });
    return change.moveNodeByKey(item.key, newSublist.key, 0);
}

function splitListItem(opts, change) {
    const { value } = change;
    const currentItem = getCurrentItem(opts, value);
    if (!currentItem) {
        return change;
    }
    const splitOffset = value.startOffset;
    return change.splitDescendantsByKey(currentItem.key, value.startKey, splitOffset);
}

function unwrapList(opts, change) {
    const items = getItemsAtRange(opts, change.value);
    if (items.isEmpty()) {
        return change;
    }
    items.forEach(item => change.unwrapNodeByKey(item.key, { normalize: false }));
    const firstItem = items.first();
    const parent = change.value.document.getParent(firstItem.key);
    let index = parent.getChildIndex(firstItem.key);
    items.forEach(item => {
        item.nodes.forEach(node => {
            change.moveNodeByKey(node.key, parent.key, index, {
                normalize: false
            });
            index += 1;
        });
    });
    items.forEach(item => change.removeNodeByKey(item.key, { normalize: false }));
    return change;
}

function wrapInList(opts, change, type, data) {
    const selectedBlocks = getHighestSelectedBlocks(change.value);
    type = type || opts.types[0];
    change.wrapBlock({
        type,
        data: slate.Data.create(data)
    }, { normalize: false });
    selectedBlocks.forEach(node => {
        if (isList(opts, node)) {
            node.nodes.forEach(({ key }) => change.unwrapNodeByKey(key, { normalize: false }));
        }
        else {
            change.wrapBlockByKey(node.key, opts.typeItem, {
                normalize: false
            });
        }
    });
    return change.normalize();
}
function getHighestSelectedBlocks(value) {
    const range = value.selection;
    const { document } = value;
    const startBlock = document.getClosestBlock(range.startKey);
    const endBlock = document.getClosestBlock(range.endKey);
    if (startBlock === endBlock) {
        return immutable.List([startBlock]);
    }
    const ancestor = document.getCommonAncestor(startBlock.key, endBlock.key);
    const startPath = ancestor.getPath(startBlock.key);
    const endPath = ancestor.getPath(endBlock.key);
    return ancestor.nodes.slice(startPath[0], endPath[0] + 1);
}

function createAPI(opts) {
    return {
        utils: {
            getCurrentItem: getCurrentItem.bind(null, opts),
            getCurrentList: getCurrentList.bind(null, opts),
            getItemDepth: getItemDepth.bind(null, opts),
            getItemsAtRange: getItemsAtRange.bind(null, opts),
            getPreviousItem: getPreviousItem.bind(null, opts),
            isList: isList.bind(null, opts),
            isSelectionInList: isSelectionInList.bind(null, opts)
        },
        changes: {
            decreaseItemDepth: bindAndScopeChange(opts, decreaseItemDepth),
            increaseItemDepth: bindAndScopeChange(opts, increaseItemDepth),
            splitListItem: bindAndScopeChange(opts, splitListItem),
            unwrapList: bindAndScopeChange(opts, unwrapList),
            wrapInList: wrapInList.bind(null, opts)
        }
    };
}
function bindAndScopeChange(opts, fn) {
    return (change, ...args) => {
        const { value } = change;
        if (!isSelectionInList(opts, value)) {
            return change;
        }
        return fn(...[opts, change].concat(args));
    };
}

class Options extends immutable.Record({
    types: ['ul_list', 'ol_list'],
    typeItem: 'list_item',
    typeDefault: 'paragraph',
    canMerge: (a, b) => a.type === b.type
}) {
}

function onBackspace(event, change, editor, opts) {
    const { value } = change;
    const { startOffset, selection } = value;
    if (selection.isExpanded) {
        return undefined;
    }
    if (startOffset > 0) {
        return undefined;
    }
    const currentItem = getCurrentItem(opts, value);
    if (!currentItem) {
        return undefined;
    }
    if (!selection.isAtStartOf(currentItem)) {
        return undefined;
    }
    event.preventDefault();
    return unwrapList(opts, change);
}

function onEnter(event, change, editor, opts) {
    if (event.shiftKey) {
        return undefined;
    }
    const { value } = change;
    const currentItem = getCurrentItem(opts, value);
    if (!currentItem) {
        return undefined;
    }
    event.preventDefault();
    if (value.isExpanded) {
        change.delete();
    }
    if (currentItem.isEmpty) {
        if (getItemDepth(opts, value) > 1) {
            return decreaseItemDepth(opts, change);
        }
        return unwrapList(opts, change);
    }
    return splitListItem(opts, change);
}

function onTab(event, change, editor, opts) {
    const { value } = change;
    const { isCollapsed } = value;
    if (!isCollapsed || !getCurrentItem(opts, value)) {
        return undefined;
    }
    if (event.shiftKey) {
        event.preventDefault();
        return decreaseItemDepth(opts, change);
    }
    event.preventDefault();
    return increaseItemDepth(opts, change);
}

const KEY_ENTER = 'Enter';
const KEY_TAB = 'Tab';
const KEY_BACKSPACE = 'Backspace';
function createPlugin(opts) {
    return {
        onKeyDown: onKeyDown.bind(null, opts)
    };
}
function onKeyDown(opts, event, change, editor) {
    const args = [event, change, editor, opts];
    switch (event.key) {
        case KEY_ENTER:
            return onEnter(...args);
        case KEY_TAB:
            return onTab(...args);
        case KEY_BACKSPACE:
            return onBackspace(...args);
        default:
            return undefined;
    }
}

function createValidation(opts) {
    return node => joinAdjacentLists(opts, node);
}
function joinAdjacentLists(opts, node) {
    if (node.object !== 'document' && node.object !== 'block') {
        return undefined;
    }
    const invalids = node.nodes
        .map((child, i) => {
        if (!isList(opts, child)) {
            return null;
        }
        const next = node.nodes.get(i + 1);
        if (!next || !isList(opts, next) || !opts.canMerge(child, next)) {
            return null;
        }
        return [child, next];
    })
        .filter(Boolean);
    if (invalids.isEmpty()) {
        return undefined;
    }
    return change => {
        invalids.reverse().forEach(pair => {
            const [first, second] = pair;
            const updatedSecond = change.value.document.getDescendant(second.key);
            updatedSecond.nodes.forEach((secondNode, index) => {
                change.moveNodeByKey(secondNode.key, first.key, first.nodes.size + index, { normalize: false });
            });
            change.removeNodeByKey(second.key, { normalize: false });
        });
    };
}

function createSchema(opts) {
    const constructedSchema = {
        validations: [createValidation(opts)],
        blocks: {
            [opts.typeItem]: {
                parent: { types: opts.types },
                nodes: [{ objects: ['block'] }],
                normalize: normalize({
                    parent_type_invalid: (change, context) => change.unwrapBlockByKey(context.node.key, {
                        normalize: false
                    }),
                    child_object_invalid: (change, context) => wrapChildrenInDefaultBlock(opts, change, context.node)
                })
            }
        }
    };
    opts.types.forEach(type => {
        constructedSchema.blocks[type] = {
            nodes: [{ types: [opts.typeItem] }],
            normalize: normalize({
                child_type_invalid: (change, context) => change.wrapBlockByKey(context.child.key, opts.typeItem, {
                    normalize: false
                })
            })
        };
    });
    return slate.Schema.create(constructedSchema);
}
function normalize(reasons) {
    return (change, reason, context) => {
        const reasonFn = reasons[reason];
        if (reasonFn) {
            reasonFn(change, context);
        }
    };
}
function wrapChildrenInDefaultBlock(opts, change, node) {
    change.wrapBlockByKey(node.nodes.first().key, opts.typeDefault, {
        normalize: false
    });
    const wrapper = change.value.document.getDescendant(node.key).nodes.first();
    node.nodes.rest().forEach((child, index) => change.moveNodeByKey(child.key, wrapper.key, index + 1, {
        normalize: false
    }));
    return change;
}

function EditList(optsInput = {}) {
    const opts = new Options(optsInput);
    return Object.assign({}, createAPI(opts), { schema: createSchema(opts), plugin: createPlugin(opts) });
}

exports.default = EditList;
//# sourceMappingURL=index.js.map
